(*i camlp4deps: "parsing/grammar.cma" i*)
(* 
Declare ML Module "explication".
*)

open Clenv
open Command;;
open Coqast;;
open Decl_kinds;;
open Declarations;;
open Declare;;
open Eauto;;
open Entries;;
open Environ
open Equality;;
open Esyntax
open Evarutil;;
open Evd;;
open Extend;;
open Genarg;;
open Inductive;;
open Inductiveops
open Libnames;;
open Metasyntax
open Nameops
open Names;;
open Pcoq;;
open Pfedit;;
open Pp;;
open Printer;;
open Proof_trees;;
open Proof_type;;
open Rawterm
open Reduction;;
open Reductionops
open Refiner;;
(* open Showproof_ct*)
open Sign;;
open Tacexpr
open Tacmach;;
open Tacred;;
open Tacticals;;
open Tactics;;
open Term;;
open Termops;;
(*open Translate*)
open Typing;;
open Util;;
open Vernacentries;;
open Vernacinterp;;

open Arbre_de_preuve

(***********************************************************************
 Quelques fonctions sur les listes
 *)
(* un tail qui ne plante pas *)
let list_tl l = match l with [] -> [] |x::l1->l1;;

(* moins les n premiers elements *)
let rec ntl l n =
  if n<=0 then l else ntl (list_tl l) (n-1)

(* les n premiers elements *)
let rec nhd l n =
  if n<=0 then [] else (List.hd l)::(nhd (list_tl l) (n-1))

let last l =
  List.hd (List.rev l)
(***********************************************************************
 Contextes, buts
*)

(* hypothese d'un but *)
let hyps pf =
  match pf with
    |Taprouver (Tcontexte h,_)->h
    |Tpreuve (Tcontexte h,_,_,_)->h
    |_->assert false

(* conclusion d'un but *)
let concl pf = 
  match pf with
    |Taprouver (h,Tterm (e,c))->c
    |Tpreuve (h,Tterm(e,c),_,_)->c
    |_->assert false

(* sous-buts d'un but *)
let sous_buts pf =
  match pf with
    |Tpreuve (h,c,_,l)->l
    |_->assert false

let tactic pf =
  match pf with
    |Tpreuve (h,c,Ttactic ((r,_),_,_),l)->r
    |_->assert false

let tactics pf =
  match pf with
    |Tpreuve (h,c,Ttactic (r,l1,l2),l)-> l1@[r]@l2
    |_-> []

let tactici pf =
   match pf with
    |Tpreuve (h,c,Ttactic (r,_,_),l)->r
    |_->assert false

let ajoute_tactic_apres t pf =
     match pf with
    |Tpreuve (h,c,Ttactic (r,l1,l2),l)-> Tpreuve (h,c,Ttactic (r,l1,t::l2),l)
    |_-> pf

let ajoute_tactic_avant t pf =
     match pf with
    |Tpreuve (h,c,Ttactic (r,l1,l2),l)-> Tpreuve (h,c,Ttactic (r,l1@[t],l2),l)
    |_-> pf

(* expression vernac d'une tactique *)
let string_of_rule r =
  pp_with  Format.str_formatter (pr_rule r);
  let s = Format.flush_str_formatter () in
    s^"."
  
(* environnement d'un  but *)
let env pf =
     match pf with
    |Taprouver (h,Tterm (e,c))->e
    |Tpreuve (h,Tterm(e,c),_,_)->e
    |_->assert false

(* sorte de t, qui est de type s *)
let sorte t s =
  if is_Prop t
  then "proposition"
  else 
    match (kind_of_term s) with
         Sort(Prop(Null)) -> "formule"
      |_ -> (match (kind_of_term t) with
                 Prod(_,_,_) -> "fonction"
               |_ -> "type")

let nom_hyp h =
  match h with
    | Thyp (Tid n,c,t)->n
    |_->assert false

let val_hyp h =
  match h with
    | Thyp (n,c,t)->c
    |_->assert false

let type_hyp h =
  match h with
    | Thyp (n,c,Tterm (e,t))->t
    |_->assert false

let type_env_hyp h =
  match h with
    | Thyp (n,c,Tterm (e,t))->e
    |_->assert false

let est_vide t =
  match t with
    |Tvide -> true
    |_->false

let env_term t =
  match t with
    |Tterm (e,_) -> e
    |_-> assert false

let val_term t =
  match t with
    |Tterm (_,t) -> t
    |_-> assert false

(* teste si un terme est une hypothese du contexte *)
let dans_contexte t lh =
  List.exists
    (fun h -> match kind_of_term t with
       | Const c -> (string_of_label (label c))= (nom_hyp h)
       | Var c ->   (string_of_id c) = (nom_hyp h)
       |_-> false)
    lh

(* teste s'il y a une hypothese de pf qui est de type t *)
let dans_contexte_type t pf=
  let env0 = env pf in
  let lterm=ref [] in
  let l=Environ.fold_named_context
	  (fun env (id,c,ty) a ->
	     let c = match c with None -> Tvide | Some c -> Tterm (env0,c) in
	     let t = Thyp (traduit_id id, c  , Tterm (env0,ty))
	     in
	       lterm:=(!lterm)@[t];
	       a)
          env0 ~init:(mt ()) in
   List.exists
    (fun (Thyp (_,_,Tterm(_,th))) -> eq_constr t th)
    !lterm

(* donne le type d'une variable d'un produit itere *)
let rec trouve_type_produit c t =
  match (kind_of_term t) with
      Prod(x,ty,t1) -> (match x with
			    Names.Name n ->
			      if n=c then ty else trouve_type_produit c t1
			  |_-> trouve_type_produit c t1)
    |_-> assert false

(* enleve les produits de tete *)
let prod_head t =
  let rec prod_head t =
    match (kind_of_term (strip_outer_cast t)) with 
	Prod(x,ty,c) ->
	  (match x with
	       Name x -> prod_head (subst1 (mkVar x) c)
	     |_-> prod_head c)
      | _ -> t
  in prod_head t (*(whd_betaiotaevar (Global.env()) Evd.empty t)*)

(* types inductifs *)
let name_ind env indf =
  let (ind,_) = dest_ind_family indf in
  let (_,mip) = lookup_mind_specif env ind in
    string_of_id mip.mind_typename
   
(* nom du constructeur i (i>=1) *)   
let  constr_of_mind env indf i =
  let (ind,_) = dest_ind_family indf in
  let (_,mip) = lookup_mind_specif env ind in
    string_of_id mip.mind_consnames.(i-1)

(* arite du constructeur i (i>=1) *)
let arity_of_constr_of_mind env indf i =
  let (ind,_) = dest_ind_family indf in
  let (_,mip) = lookup_mind_specif env ind in
  let p = mip.mind_nparams in
    (get_constructors env indf).(i-1).cs_nargs
    
(* t est le meme type inductif que ti *)
let  hd_is_mind env t ti =
    try (let IndType (indf,targ) = find_rectype env Evd.empty t in
	 (name_ind env indf) = ti)
    with _ -> false

(* arguments du constructeur c (c>=1): recursifs ou non 
   on oublie les parametres *)
let mind_ind_info_hyp_constr env indf c =
  let ti = name_ind env indf in
  let (ind,params) = dest_ind_family indf in
  let nparams = List.length params in
  let lp= ntl (get_constructors env indf).(c-1).cs_args nparams in
    List.map (fun (_,_,t)-> if hd_is_mind env t ti
	      then "argrec" else "arg")
      lp
(***********************************************************************
Tactiques particulieres
*)

let atom_tactic pf =
  let tac = tactic pf in
    match tac with
      | Tactic(TacAtom (_,tac1),_) -> tac1
      |_-> assert false
	 
let est_generalize pf =
  try (match atom_tactic pf with
	 |  TacGeneralize _ ->true
	 |_->false)
  with _->false

let est_pose pf =
  try (match atom_tactic pf with
	 |  TacForward (true, Names.Name x, c) -> true
	 |_->false)
  with _->false

let est_assert pf =
  try (match atom_tactic pf with
	 |  TacForward (false, Names.Name x, c) -> true
	 |_->false)
  with _->false

let est_cut pf =
  try (match atom_tactic pf with
	 |  TacCut _ -> true
	 |_->false)
  with _->false

let est_absurde pf =
  let tac = tactic pf in
    (match tac with
       | Tactic(TacAtom (_,tac1),_) ->
	   (match tac1 with 
	      | TacExtend (_,"Absurd",_) -> true
	      |_->false)
       |_-> false)
    
let est_intros pf =
  let r = tactic pf in
    (match r with
       | Tactic (TacArg (Tacexp (TacAtom (_,TacIntroPattern _))),_)
       | Tactic (TacAtom (_,TacIntroPattern _),_)
       | Tactic (TacArg (Tacexp (TacAtom (_,TacIntroMove _))),_)
       | Tactic (TacAtom (_,TacIntroMove _),_) -> true
       |_-> false)
    
let est_trivial pf =
  let tac = tactic pf in
	(match tac with
	   | Tactic(TacAtom (_,tac1),_) ->
	       (match tac1 with 
		  | TacTrivial _ | TacAssumption -> true
		  |_->false)
	   |_-> false)

let est_unfoldin pf =
  let tac = tactic pf in
	(match tac with
	   | Tactic(TacAtom (_,tac1),_) ->
	       (match tac1 with 
		  | TacReduce (Rawterm.Unfold [([], _)],
			       [Tacexpr.InHyp id]) -> true
		  |_->false)
	   |_-> false)

let est_unfold pf =
   let tac = tactic pf in
	(match tac with
	   | Tactic(TacAtom (_,tac1),_) ->
	       (match tac1 with 
		  | TacReduce (Rawterm.Unfold [([], _)],
			       []) -> true
		  |_->false)
	   |_-> false)

let est_simpl pf = (* frederique *)
   let tac = tactic pf in
	(match tac with
	   | Tactic(TacAtom (_,tac1),_) ->
	       (match tac1 with
			| TacReduce (Rawterm.Red false, [])-> false (* Red *)
	   		| TacReduce (Rawterm.Unfold [([], _)],[]) -> false
	   		| TacReduce (Rawterm.Unfold [([], _)],_)-> false
			| TacReduce (r,cl)  -> true 
			|_->false)
	   |_-> false)


let est_rewritein pf =
   let tac = tactic pf in
	(match tac with
	   | Tactic(TacAtom (_,tac1),_) ->
	       (match tac1 with 
		  | TacExtend (_,"RewriteIn",_) -> true
		  |_->false)
	   |_-> false)

let est_rewrite pf =
  let tac = tactic pf in
	(match tac with
	   | Tactic(TacAtom (_,tac1),_) ->
	       (match tac1 with 
		  | TacExtend (_,"Rewrite",_) -> true
		  |_->false)
	   |_-> false)

let est_elim pf = (* frederique *)
  let tac = tactic pf in
	(match tac with
	   | Tactic(TacAtom (_,tac1),_) ->
	       (match tac1 with 
		  | TacElim _ -> true   
		  |_->false)
	   |_-> false)

let est_recurrence pf =
  let tac = tactic pf in
	(match tac with
	   | Tactic(TacAtom (_,tac1),_) ->
	       (match tac1 with 
		  | TacOldInduction _
		  | TacNewInduction _
		  | TacCase _
		  | TacElim _ -> true  
		  |_->false)
	   |_-> false)


(**********************************************************************
Arguments (constr) d'une tactique
*)

let string_to_constr pf x =
  Constrintern.interp_constr Evd.empty (env pf) (Topconstr.CRef (Ident ((0,0), id_of_string x)))
    
let id_to_constr pf x =
   Constrintern.interp_constr Evd.empty (env pf) (Topconstr.CRef (Ident ((0,0), x)))

let rec rule_arguments tac pf =
  let Ttactic ((tac,_),_,_) = tac in
  try(
  match tac with
    | Tactic(TacArg (TacCall (_, Ident (_, idtac), l)),_) ->
	(match (string_of_id idtac) with
	     _ -> List.map (fun (ConstrMayEval (ConstrTerm c)) -> c) l
	)
    | Tactic(TacAtom (_,tac1),_) ->
	(match tac1 with 
	  | TacSplit(ImplicitBindings l) -> l
	  | TacGeneralize l -> l (* l= 'constr list *)
	  | TacRight l -> []
	  | TacLeft l -> []
	  | TacReduce (Rawterm.Red false, [])-> [] (* Red *) 
	  | TacReduce (Rawterm.Unfold [([], EvalConstRef a)],_(*[Tacexpr.InHyp id]*))->
	      let c= Names.string_of_label (Names.label a) in
		[string_to_constr pf c]
		(*
	  | TacReduce (Rawterm.Unfold [([], a (*(Ident (_, a)) *))],[])->
	      let  Rawterm.AN a = Obj.magic a in
	       let c = Libnames.string_of_qualid (snd (qualid_of_reference a)) in
		[string_to_constr pf c]
		*)
	  | TacReduce (r,cl)  -> [] (* "Simpl" ???*)
	  | TacAuto (no,lo) ->
	      (match lo with
		   None -> []
		 |Some l -> List.map (string_to_constr pf) l)
	  | TacTrivial _ -> []
	  | TacAssumption -> []
	  | TacClear l -> [] (* identifier or_metanum list *)
	  | TacOldInduction (NamedHyp c) ->
	      (try (let tc= trouve_type_produit c (concl pf) in
		      [mkVar c;tc])
	       with _-> [id_to_constr pf c])
	  | TacNewInduction (ElimOnIdent (_, c), None, []) ->
	      (try (let tc= trouve_type_produit c (concl pf) in
		      [mkVar c;tc])
	       with _-> [id_to_constr pf c])
	  | TacApply (c,_) ->  [c]
	  | TacExact c -> [c]
	  | TacCut c -> [c]
	  | TacTrueCut (None, c)-> [] (* Assert *) 
	  | TacCase (c,_) -> [c]
	  | TacElim ((c,_),_) -> [c]
          | TacForward (true, Names.Name x, c) -> [] (* Pose *)
          | TacForward (false, Names.Name x, c) -> [c] (* Assert H:=(lemme ...)Pose *)
	  | TacExtend (_,"InfoAuto",[]) -> []
	  | TacExtend (_,"EAuto",_) -> []
	  | TacExtend (_,"Rewrite",[_;a]) ->
	      let (c,_) = out_gen wit_constr_with_bindings a in [c]
	  | TacExtend (_,"RewriteIn", [_;a; id]) ->
	      let (c,_) = out_gen wit_constr_with_bindings a in
	      (* let (id,_) = out_gen wit_constr_with_bindings id in
		 out_gen plante a l'execution... *)
		[c (*; id *)]
	  |_-> [])
    |_->[])
  with _ -> []

(**********************************************************************
  Modifs de l'arbre de preuve
*)

let est_fleche pf =
  let cl = concl pf in
    match (kind_of_term cl) with
	Prod(Anonymous,_,_) ->  true
      |_-> false
	 
(* Virer les buts Intros:
 - apres Generalize, Cut
 - apres Elim/Induction, si le but est une fleche  i.e. un produit non dependant
*)
let virer_intros1 b pf =
  if b && (est_intros pf)
  then ajoute_tactic_avant (tactici pf) (List.hd (sous_buts pf))
 
 else if est_generalize pf
  then (let pf1=List.hd (sous_buts pf) in
	  if est_intros pf1
	  then let Tpreuve (h,c,r,_) = pf in
	    ajoute_tactic_apres (tactici pf1) (Tpreuve (h,c,r,sous_buts pf1))
	  else pf)
  
else if est_cut pf
  then (print_preuve pf;
        print_string ((string_of_int (List.length (sous_buts pf)))^" sous-buts\n");
        let [pf1;pf2]= sous_buts pf in
	  let pf1 = if est_intros pf1
	            then ajoute_tactic_apres (tactici pf1) (List.hd (sous_buts pf1))
                    else pf1 in
          let Tpreuve (h,c,r,_) = pf in
	    Tpreuve (h,c,r,[pf2;pf1]))

  else if est_recurrence pf
  then (let lpf = List.map (fun pf1 ->
			      if est_intros pf1 && est_fleche pf1
			      then ajoute_tactic_avant (tactici pf1) (List.hd (sous_buts pf1))
			      else pf1)
		    (sous_buts pf) in
	  let Tpreuve (h,c,r,_) = pf in
	    Tpreuve (h,c,r,lpf))
  else pf

(* Virer les buts avant Intros, si le but est une fleche
   i.e. un produit non dependant
*)
let rec virer_intros b pf1 =
  let pf = virer_intros1 b pf1 in
  if pf<>pf1 && b=true then virer_intros false pf
  else
    match pf with
      | Taprouver _ -> pf
      | Tpreuve (h,c,r,lpf) ->
	  Tpreuve (h,c,r,List.map (virer_intros false) lpf)
      |_->assert false

(***********************************************************************
 Arbre de texte
*)

type texte =
  |Trien
  |Tsp (* espace *)
  |Tnl (* retour a la ligne *)
  |Tstr of string (* chaine *)
  |Tnom of string (* chaine *)
  |Tph of texte (* Phrase complete *)
  |Tlist of texte list (* boite hov *)
  |Tseq of  preuve (* contexte *)
  |Tterme of env * constr (* terme *)
  |Tindent of texte list (* indentation *)
  |Top of string * (texte list) (* string: type de liste *)
  |Ttac of (rule * int) list
      
let pr_concl pf = Tterme ((env pf),(concl pf))

let pr_list sep f l =
  Top (sep,List.map f l)

let pr_fnl = "nl"

let prl_nl = pr_list "nl"

let pr_term env c = Tterme (env,c)

(* elimine les buts repetes *)
let ens_pf lpf=
  let res = ref [] in
    List.iter (fun t ->
		 if (not (List.exists
			    (fun a ->
			       (eq_constr (concl a) (concl t)))
			    (!res)))
		 then  res:=(!res)@[t])
      lpf;
    !res

let prlt_nl lpf =
  let lpf = ens_pf lpf in
  pr_list ",et"
    (fun pf1 ->
       pr_term (env pf1) (concl pf1))
    lpf

(* pas de separateur a la fin *)
let pr_list1 sep f l = pr_list sep f l 

let t_l1 s a = Top (s,[a])
let t_l2 s a b = Top (s,[a;b])
let t_l3 s a b c= Top (s,[a;b;c])
let t_l4 s a b c d= Top (s,[a;b;c;d])


(*
taille d'un constr.
*)
let string_of_constr e c =
  pp_with  Format.str_formatter (prterm_env_at_top e c);
  Format.flush_str_formatter ()

let string_of_rule r =
  pp_with  Format.str_formatter (pr_rule r);
  Format.flush_str_formatter ()
  
let taille e c =
  String.length (string_of_constr e c)

let string_of_stream p =
  pp_with  Format.str_formatter p;
  Format.flush_str_formatter ()

(***********************************************************************
 Textes modeles
*)

let t_demontrons_A pf =
  t_l1 "Dmontrons_#A_:" (pr_concl pf)

let t_a_prouver hyps pf =
   t_l1 "_prouver:_#A" (Tseq hyps)

let t_est_clair cl0 pf =
  t_l1 "#A_est_clair" (pr_term (env pf) cl0)

let t_donc_A a =
  t_l1 "donc_#A_," a

let t_Donc_A a =
  t_l1 "Donc_#A_," a

let t_De_A_on_deduit_B a b=
   t_l2 "De_#A_on_dduit_#B" a b

let t_De_A_on_deduit_B_par_definition a b=
   t_l2 "De_#A_on_dduit_#B_par_dfinition" a b

let t_De_A_type_B_on_deduit_C a b c =
  t_l3 "De_#A_:_#B_on_dduit_#C" a b c

let t_ainsi_A_B a b =
  match a with
    | Trien -> t_l1 "Ainsi_#A" b
    | _ -> t_l2 "Ainsi_,_#A_,_#B" a b

let t_De_plus_A a =
  t_l1 "De_plus_#A" a

let t_Pour_prouver_A_B_procedons_ainsi a b=
  t_l2 "Pour_prouver_#A_,_#B_procdons_ainsi:" a b

let t_enumerer f l =
  Top ("numrer", List.map f l)

let t_A_grace_a_B a b =
  t_l2 "#A_grce__#B" a b

let t_par_A_on_a_B a b=
  match a with
    |Trien -> t_l1 "On_a_#B" b
    |_-> t_l2 "Par_#A_on_a_#B" a b

let t_distinguons_N_cas_pour_A n a =
  t_l2 "Distinguons_#N_cas_pour_#A" (Tstr (string_of_int n)) a

let t_par_recurrence_sur_la_definition_de_A a=
  t_l1 "Par_rcurrence_sur_la_dfinition_de_#A" a

let t_Ainsi_dans_tous_les_cas_on_a_montre_A a =
  t_l1 "Ainsi,_dans_tous_les_cas_on_a_dmontr_#A" a

let t_Ainsi_dans_tous_les_cas_de_A_on_deduit_B a b =
  t_l2 "Ainsi,_dans_tous_les_cas,_de_#A_,_on_dduit_#B" a b

let t_De_la_definition_de_A_de_B_et_de_C_on_deduit_D a b c d=
  t_l4 "De_la_dfinition_de_#A_,_de_#B_,_et_de_#C_,_on_dduit_#D" a b c d

let t_par_definition_de_A a =
  t_l1 "par_dfinition_de_#A" a

let t_Procedons_par_recurrence_sur_A_de_type_B a b =
  t_l2 "Procdons_par_rcurrence_sur_#A_(de_type_#B_)" a b

let t_cas_de_recurrence () =
  Tstr "- Cas de rcurrence"

let t_cas_de_base ()=
  Tstr "- Cas de base"

let t_grace_a_A intro a=
  match intro with
      "par" -> t_l1 "Par_#A" a
    |"grace a"|_-> t_l1 "grce__#A" a

let t_Par_l_egalite_A a =
  t_l1 "par_l'galit_#A" a


let t_on_obtient_A a =
  t_l1 "on_obtient_#A" a

let t_On_obtient_A a =
  t_l1 "On_obtient_#A" a

let t_on_peut_supposer_A a =
  t_l1 "on_peut_supposer_#A" a

let t_Soit_A a =
  t_l1 "Soit_#A" a

let t_On_a_A a =
  t_l1 "On_a_#A" a

let t_A_partir_de_A_type_B a b =
  t_l2 "_partir_de_#A_:_#B" a b

let t_A_partir_de_A_type_B_on_deduit_C a b c =
  t_l3 "_partir_de_#A_:_#B_on_dduit_#C" a b c

let t_On_obtient_une_contradiction_avec_A_et_B a b =
  t_l2 "On_obtient_une_contradiction_avec_#A_et_#B" a b

let t_constructeur_A a =
  t_l1 "constructeur_#A" a

let t_Avec_A_type_B_on_obtient_une_contradiction a b =
  t_l2 "Avec_#A_:_#B_on_obtient_une_contradiction" a b

let t_De_A_et_B_type_C_on_obtient_une_contradiction a b c =
  t_l3 "De_#A_et_#B_:_#C_on_obtient_une_contradiction" a b c

let t_Soit n =
  if n>1 then "Soient " else "Soit "

let t_une_proposition n =
  if n>1 then " des propositions" else " une proposition"

let t_telle_que n =
  if n>1 then " telles que " else " telle que "

let t_tel_que n =
  if n>1 then " tels que " else " tel que "

let t_un_element n =
  if n>1 then " des lments de " else " un lment de "

let t_Supposons () =
  "Supposons "

let t_et () =
  "et"

(***********************************************************************
Explication des preuves 
*)

(* Parametres globaux *)
(* expliquer les arguments des tactiques *)
let p_grace_a = ref true;;
(* donner les enonces des hypotheses et propositions utilisees  *)
let p_type = ref true;;

let exp_type e x tx =
  if !p_type
  then let stx = taille e tx in
    if stx < 50 then [pr_term e x; Tstr " : "; pr_term e tx (* Tstr (string_of_int stx) *)]
    else [pr_term e x]
  else [pr_term e x]

let debug = ref false;;

let deb s = if !debug then Tstr s else Tstr ""

(* Debut de l'explication *)
let rec expliquer0 pf =
  let hyps' = hyps pf in 
    Tlist [if hyps'<> []
	  then expliquer_hyps "soit"  hyps' pf
	  else Trien;
	  t_demontrons_A pf;
	  Tindent [expliquer  pf]]

(* explication d'un but *)
and expliquer  pf =
  let cl0=(concl pf) in
  let hyps' = (hyps pf) in
  let exptac= Ttac (tactics pf) in
    (* on met la tactique au debut si les sous-buts ont des hypotheses nouvelles *)
  let res=
    match pf with
      | Taprouver (h,c)-> t_a_prouver h pf
      | Tpreuve (h,c,r,spfl) ->
	  if est_recurrence pf
	  then Tlist [exptac; expliquer_recurrence r  cl0 pf spfl]
	  else (match spfl with
	     | [] -> Tlist [exptac; expliquer_but_prouve r cl0 pf]
	     | [pf1] ->
		 let delta = (hyps pf1) in
		 let cl1=(concl pf1) in
		   if (delta=[])&&(eq_constr cl1 cl0)
		   then Tlist [exptac; expliquer_cas0 r  cl0 pf  cl1 pf1 delta]
		   else if (delta<>[])&&(eq_constr cl1 cl0)
		   then Tlist [exptac; expliquer_cas1 r  cl0 pf  cl1 pf1 delta]
		   else if (delta=[])&&(not (eq_constr cl1 cl0))
		   then Tlist [expliquer_cas2 r  cl0 pf  cl1 pf1 delta; exptac]
		   else  Tlist [exptac; expliquer_cas3 r  cl0 pf  cl1 pf1 delta]
		     
	     |_-> (* recherche du but cl0 dans les sous-buts *)
		let (l1,lb) = cherche_but cl0 spfl in
		  (match lb with
		     | []->  Tlist [exptac;expliquer_cas6 r  cl0 pf spfl]
		     |[b] ->  Tlist [exptac; expliquer_cas5 r  cl0 pf l1 b]
		     |_ -> Tlist [exptac; expliquer_cas4 r  cl0 pf l1 lb]
		  ))
  in res 

(* H |- G <= H |- G' *)
and est_cas2 osign pf =
  let cl0=(concl pf) in
    let r = tactic pf in
    let spfl = sous_buts pf in
	  (match spfl with
	     | [pf1] ->
		 let hyps1 = (hyps pf1) in
		 let cl1=(concl pf1) in
		   (hyps1=[])&&(not (eq_constr cl1 cl0))
	     |_->false)
	
(***********************************************************************
Les cas qui ne dependent pas de la tactique utilisee
*)
and expliquer_but_prouve r cl0 pf =
  if (est_trivial pf) || (* Trivial, Assumption *)
    (dans_contexte_type cl0 pf) (* le but est dans les hypotheses *)
  then  Trien
  else
    (match rule_arguments r pf with
	 [x]->
	   if dans_contexte x (hyps pf) (* l'argument de la tactique est une hypothese*)
	   then t_A_grace_a_B
	     (t_De_A_on_deduit_B  (pr_term (env pf) x)
		(pr_term (env pf) cl0))
	     Trien
	   else t_A_grace_a_B
	     (t_De_A_type_B_on_deduit_C (pr_term (env pf) x)
		(pr_term (env pf) (type_of (env pf) Evd.empty x))
		(pr_term (env pf) cl0))
	     Trien
       |_-> Tph (t_est_clair cl0 pf))
    
(* Cas 0:
	  H |- G <= H |- G
Auto
*)
and expliquer_cas0 r  cl0 pf  cl1 pf1 delta =
  Tlist [expliquer pf1]

(* Cas 1:
         H |- G <= H,D |- G
Pose, Generalize;Intros
*)
and expliquer_cas1 r  cl0 pf  cl1 pf1 delta =
  if est_pose pf
  then Tlist ([expliquer_hyps "soit"  delta pf1]@
	     [expliquer  pf1])
  else if est_assert pf
  then Tlist [Tph (t_A_grace_a_B (expliquer_hyps "on a"   delta pf1)
		     (grace_a "grace a" r pf pf1));
	      expliquer  pf1]
  else if est_rewritein pf
  then Tlist [Tph (t_A_grace_a_B (expliquer_hyps "on obtient"   delta pf1)
		     (grace_a "grace a" r pf pf1));
	      expliquer  pf1]
  else if est_unfoldin pf
  then Tlist [Tph (t_A_grace_a_B (expliquer_hyps "on obtient"  delta pf1)
		     (grace_a "grace a" r pf pf1));
	      expliquer pf1]
  else  Tlist [Tph (Tlist [grace_a "par" r pf pf1;Tstr ",";
			   (expliquer_hyps "on peut supposer" delta pf1)]);
	      expliquer pf1]
    
(* Cas 2:
         H |- G <= H |- G' 
 Apply, Rewrite
*)
and expliquer_cas2 r  cl0 pf  cl1 pf1 delta =
	if est_rewrite pf (* frederique *)
 	then (let [a] = (rule_arguments r pf) in
		let ta = (type_of (env pf) Evd.empty a) in
	
		let a'=prod_head (type_of (env pf) Evd.empty a) in

	 	 Tlist [expliquer pf1;
		  	(t_De_A_on_deduit_B 
			(Top (",et", [(pr_term (env pf1) cl1);	
			(pr_term (env pf) ta)]))
	(* ceci donne H : type de H   ->  (Tlist (exp_type (env pf) a ta))*)
	(* ceci donne H               ->  (pr_term (env pf) a) *) 
			(pr_term (env pf) cl0))])

	 else if est_simpl pf (* frederique *)
 	then (Tlist [expliquer pf1;
		  	(t_De_A_on_deduit_B_par_definition
			     (pr_term (env pf1) cl1)
			     (pr_term (env pf) cl0))])

    else  Tlist [expliquer pf1;
	   (t_A_grace_a_B (t_De_A_on_deduit_B
			     (pr_term (env pf1) cl1)
			     (pr_term (env pf) cl0))
	      (grace_a "grace a" r pf pf1))]
    
(* Cas 3:
   H |- G <= H,D |- G' *) 
and expliquer_cas3 r  cl0 pf  cl1 pf1 delta =
  Tlist [expliquer_hyps "soit"  delta pf1;
	Tindent [expliquer pf1];
	Tph (t_ainsi_A_B (grace_a "grace a" r pf pf1)
	       (pr_term (env pf) cl0))] 
    

(* Cas 6: on n'a pas le but dans les sous-buts
   H |- G <= H1 |- G1,..., Hn |- Gn *)
and expliquer_cas6 r  cl0 pf spfl =
  if est_rewrite pf (* merdique *)
  then (let [a] = (rule_arguments r pf) in
	let a'=prod_head (type_of (env pf) Evd.empty a) in
	  Tlist [expliquer_lpf  spfl;
	(t_A_grace_a_B (t_De_A_on_deduit_B
				       (prlt_nl spfl)
				       (pr_term (env pf) cl0))
			(grace_a "grace a" r pf pf))])

  else if est_absurde pf
  then let pf1=List.hd spfl in
  let pf2 = List.nth spfl 1 in
    Tlist [expliquer_lpf  spfl;
	   Tph (t_On_obtient_une_contradiction_avec_A_et_B
		  (pr_term (env pf1) (concl pf1))
		  (pr_term (env pf2) (concl pf2)))]
  else Tlist [expliquer_lpf  spfl;
	      (t_A_grace_a_B (t_De_A_on_deduit_B
				    (prlt_nl spfl)
				    (pr_term (env pf) cl0))
		     (grace_a "grace a" r pf pf))]
      
(* Cas 5: le but apparait une fois dans les sous-buts
   H |- G <= H1 |- G1,..., Hn |- Gn  et H,D |- G 
(Cut ...;Intros) *)
(* frederique : il n'y a pas de grace_a aurait-on oublier des cas ?*)
and expliquer_cas5 r  cl0 pf l1 b =
  Tlist [expliquer_lpf  l1;
	expliquer_hyps_but "on a"  b]
    
(* Cas 4: le but apparait plusieurs fois dans les sous-buts
   H |- G <= H1 |- G1,..., Hn |- Gn
   et H'1,D1 |- G ... H'm,Dm |- G *)
and expliquer_cas4 r  cl0 pf l1 lb =
  Tlist [expliquer_lpf  l1;
	if l1<>[]
	 then Tlist [t_Pour_prouver_A_B_procedons_ainsi
		      (pr_term (env pf) cl0)
		      (grace_a "grace a" r pf pf)]
	else (*demonstration par cas*) grace_a "grace a" r pf pf;
	t_enumerer (expliquer_hyps_but "soit" ) lb]


(***********************************************************************
Recurrences
*)

(*
Declare ML Module "explication".
Drop.
#use "base_include";;
#trace Inductiveops.find_rectype;;
go();;
*)
(* Elimination de x, le type tx de x est un produit avec un symbole de tete qui est un type inductif P:.
  tx = T1->...->Tn->(P a1...am)
*)
and expliquer_recurrence r  cl0 pf spfl =
  let lx=rule_arguments r pf in
  let e= env pf in
  let (x,tx) = (match lx with (* plante avec Induction 1 *)
		[x]-> x,type_of e Evd.empty x
	      |[x;tx] -> x,tx) in
  let (_,tx) = reduce_to_quantified_ind e Evd.empty tx in
  let sx = type_of e Evd.empty tx in
  let txind = prod_head  tx in
  let IndType (indf,targ) = Inductiveops.find_rectype e Evd.empty txind in
  let nc= Array.length(get_constructors e indf) in
  let lpfrec= nhd spfl nc in (* les buts de la recurrence *)
  let lpf = ntl spfl nc in (* les premisses de tx *)
    (* les sous-but avec un but different de cl0 *)
  let spfl'=List.filter (fun pf1 -> not (eq_constr (concl pf1) cl0)) spfl in
  let premisses =  expliquer_lpf  lpf in
    match sorte tx sx with
      | "formule" -> 	(* terme logique: dans Prop *)
	  if nc = 1
	  then (* un constructeur (/\, ex, eqT) *)
	    let pf1 = virer_intros1 true (List.hd lpfrec) in
	    let delta = (hyps pf1) in
	    let cl1=(concl pf1) in
	      if (eq_constr cl1 cl0) (* meme but *)
	      then Tlist [premisses;
			  let avant = t_A_partir_de_A_type_B (pr_term e x) (pr_term e tx) in
			    expliquer_hyps_but_avant avant "on obtient"  pf1]
	      else (* but change *)
		Tlist [premisses;
		       expliquer_lpf  lpfrec;
		       Tph (t_A_partir_de_A_type_B_on_deduit_C
			      (pr_term e x) (pr_term e tx) (pr_term e cl0))]
	  else if nc=0 (* elimination de False *)
	  then Tlist [premisses;
		      if lpf=[]
		      then (Tph (t_Avec_A_type_B_on_obtient_une_contradiction (pr_term e x) (pr_term e tx)))
		      else (Tph (t_De_A_et_B_type_C_on_obtient_une_contradiction
				  (prlt_nl lpf)
				  (pr_term e x) (pr_term e tx)))]
	  else (* plusieurs constructeurs (\/ a 2 constructeurs *)
	    let argrec=ref false in
	       for i=1 to nc do
		 if List.mem "argrec" (mind_ind_info_hyp_constr e indf i)
		 then argrec:=true;
	       done;
	       Tlist [premisses;
		     if !argrec = false
		     then Tph (t_distinguons_N_cas_pour_A nc (pr_term e txind))
		     else Tph (t_par_recurrence_sur_la_definition_de_A
				  (pr_term e txind));
		     t_enumerer (expliquer_hyps_but "soit" ) lpfrec;
		     if true (* lpf=[]*)
		     then (Tph (t_Ainsi_dans_tous_les_cas_on_a_montre_A
				  (pr_term e cl0)))
		     else (Tph (t_Ainsi_dans_tous_les_cas_de_A_on_deduit_B
				  (prlt_nl lpf)
				  (pr_term e cl0)))]

      |_-> (* terme calculatoire: Type ou Set *)
	 if nc = 1
	 then (* un seul constructeur *)
	    Tlist [premisses;
		  expliquer_lpf  lpfrec;
		  Tph (t_De_la_definition_de_A_de_B_et_de_C_on_deduit_D
			 (pr_term e x)
			 (prlt_nl lpf)
			 (prlt_nl lpfrec)
			 (pr_term e cl0))]
	 else (* plusieurs constructeurs *)
	   let c=ref 0 in 
	   Tlist [premisses;
		 (t_Procedons_par_recurrence_sur_A_de_type_B
			 (pr_term e x)
			 (pr_term e tx));
		 (prl_nl 
			 (fun pf1 -> c:=(!c)+1;
			    if (!c)>nc
			    then Tindent [
				       expliquer_hyps_but "soit"  pf1]
			    else Tindent [
			      (if List.mem "argrec" (mind_ind_info_hyp_constr e indf !c)
			       then t_cas_de_recurrence ()
			       else t_cas_de_base ()); Tsp;
			      t_constructeur_A (Tstr (constr_of_mind e indf !c));
			      expliquer_hyps_but "soit"  pf1])
		    spfl);
		 Tph (t_De_A_on_deduit_B (prlt_nl spfl) (pr_term e cl0))]
	     

(* Explication des tactiques avec arguments *)
and grace_a intro tac pf pf1 =
  if !p_grace_a =false
  then Trien
  else
  let la = rule_arguments tac pf1 in
  let defaut = (if la = []
		then Trien
		else (t_grace_a_A intro
			(pr_list1 ","
			   (fun a ->
			      let ta = (type_of (env pf1) Evd.empty a) in
			      let sa = (type_of (env pf1) Evd.empty ta) in
				match sorte ta sa with
				  |"formule" -> 
				     Tlist (exp_type (env pf1) a ta)
				  |_-> pr_term (env pf1) a)
			   la))) in
    if est_unfoldin pf && la<>[] (* TODO traiter le cas avec plusieurs el dans la *)
    then let a::_=la in t_par_definition_de_A (pr_term (env pf1) a)
    else if est_unfold pf && la<>[] (* TODO traiter le cas avec plusieurs el dans la *)
    then let a::_=la in t_par_definition_de_A (pr_term (env pf1) a)
    else
      let Ttactic ((tac,i),_,_) = tac in
    match tac with
      | Tactic(TacArg (TacCall (_, Ident (_, idtac), l)),_) ->
	  (match (string_of_id idtac) with
	       _ -> defaut
	  )
      | Tactic(TacAtom (_,tac1),_) ->
	  (match tac1 with 
	     | TacSplit(ImplicitBindings l) ->
		 Top ("en_choisissant_#A", List.map (fun a -> pr_term (env pf1) a) la)
	     | TacOldInduction (NamedHyp id) -> defaut
	     | TacNewInduction (ElimOnIdent (_, c), None, []) ->  defaut
	     | TacCase (c,_) -> defaut
	     | TacElim ((c,_),_) -> defaut
(*	     | TacExtend (_,"RewriteIn", [_;a; id]) ->
		 let [a]=la in
		   [Tlist (t_Par_l_egalite_A
			    (let ta = (type_of (env pf1) Evd.empty a) in
			       pr_term (env pf1) ta));
		    Tnl]*)
	     |_-> defaut)
      |_-> defaut
	 
and cherche_but cl spfl =
  let l1 = ref [] in
  let lb = ref [] in
    List.iter (fun pf -> if (eq_constr (concl pf) cl)
	       then lb:=(!lb)@[pf]
	       else l1:=(!l1)@[pf])
      spfl;
    (!l1,!lb)
    
and expliquer_lpf  lpf =
  let lpf = ens_pf lpf in
  let le = List.map (fun pf1 -> Tindent [expliquer_hyps_but "soit"  pf1]) lpf in
  let le = List.filter (fun e -> match e with
                         |Tindent [Tlist [_;Trien]] -> false (* attention on perd les tactiques *)
                         |_-> true)
                       le in
  Top ("nl",le)

and expliquer_hyps intro_hyps  hyps pf =
  match intro_hyps  with
    |"on a"-> Tph (t_On_a_A (texte_contexte  hyps))
    |"on obtient"-> (t_On_obtient_A (texte_contexte   hyps))
    |"on peut supposer"-> (t_on_peut_supposer_A (texte_contexte hyps))
    |"soit"|_-> expliquer_contexte  hyps

and expliquer_hyps_but intro_hyps  pf =
  let delta = (hyps pf) in
    if delta = []
    then expliquer pf
    else Tlist [expliquer_hyps intro_hyps   delta pf;
	       expliquer  pf]

and expliquer_hyps_avant av intro_hyps  hyps pf =
  Tph (Tlist [av;
	      match intro_hyps  with
		|"on a"->(t_On_a_A (texte_contexte   hyps))
		|"on obtient"-> (t_on_obtient_A (texte_contexte  hyps))
		|"on peut supposer"->(t_on_peut_supposer_A (texte_contexte hyps))
		|"soit"|_->(t_Soit_A (texte_contexte   hyps))])

and expliquer_hyps_but_avant a intro_hyps  pf =
  let delta = (hyps pf) in
    if delta = []
    then expliquer pf
    else Tlist [expliquer_hyps_avant a intro_hyps  delta  pf;
	       (expliquer pf)]

(***********************************************************************
 Explication des contextes 
*)
and group_hyps hyps =
  let lgroup=ref [] in
  let gsorte=ref "" in
  let gtype = ref None in
  let lterm=ref [] in
  let l=List.iter 
	  (fun h ->
	     let id = nom_hyp h in
	     let c=val_hyp h in
	     let e=type_env_hyp h in
	     let ty = type_hyp h in
	     let s= sorte ty (type_of e  Evd.empty ty) in
	       (match !gsorte with
		    "" -> gsorte:=s;gtype:=Some ty;
		      lterm:=[(id,c,e,ty)]
		  |sg -> if sg=s && (s="proposition"|| s="formule" || (eq_constr (let Some gt = (!gtype) in gt) ty))
		    then lterm:=(!lterm)@[(id,c,e,ty)]
		    else (lgroup:=(!lgroup)@[!gsorte,!lterm];
			  lterm:=[id,c,e,ty];
			  gsorte:=s;
			  gtype:=Some ty))
	  )
          hyps in
    lgroup:=(!lgroup)@[!gsorte,!lterm];
    (!lgroup)
      
and pr_id id = Tnom id
		 
and expliquer_contexte hyps =
  let lgroup=group_hyps hyps in
  let avant = ref "" in
  let t_avant = ref Trien in
  let res = ref [] in
    List.iter
      (fun (s,lh)->
	 let n=List.length lh in
	   match s with
	     |"proposition" ->
		let r= Tlist [Tstr (!avant); Tsp;
			      Tstr (t_Soit n);Tsp;
			      (pr_list ",et"
				 (fun (id,c,e,t)->
				    if est_vide c
				    then pr_id id
				    else Tlist [pr_id id; Tstr ":="; Tterme (env_term c) (val_term c)])
				 lh);
			      Tstr (t_une_proposition n)] in
		   avant:= t_telle_que n;
		  t_avant:=r
	     |"formule"->
		let tr = (pr_list ",et" 
			    (fun (id,c,e,t)->
			       Top("hypothese",[Tterme e t; Tstr " ("; pr_id id; Tstr ")"]))
			    lh) in
		let r = (if !t_avant=Trien
			 then Tph (Tlist [Tstr (t_Supposons ()); tr])
			 else Tph (Tlist [!t_avant; Tstr (!avant);Tsp; tr])) in
		  res:=(!res)@[r];
		  avant:="";
		  t_avant:=Trien
	     |"type"|"fonction"->
		let (_,_,e,t)=List.hd lh in
		let r = Tlist ((if !t_avant=Trien
				then []
				else [Tph !t_avant;Tsp])@
			       [Tstr (t_Soit n);Tsp;
				(pr_list ",et" 
				   (fun (id,c,e,t)->
				      if est_vide c
				      then pr_id id
					else Tlist [pr_id id;Tstr ":=";
							    Tterme (env_term c) (val_term c)])
				  lh);
				Tstr (t_un_element n);Tsp;
				Tterme e t]) in
		  avant:=t_tel_que n;
		  t_avant:=r)
      lgroup;
    if !t_avant<>Trien
    then res:=(!res)@[Tph (!t_avant)];
    Tlist (!res)

(*
 ~/pcoq/V7/bin/coqtop.byte.exe
Declare ML Module "explication".
 Goal (a,b:nat)a=b->(b:bool)b=b->b=b.
Intros a b H0 n H.
Exp.
*)

and texte_contexte hyps =
  let lterm=ref [] in
  let l=List.iter
	  (fun h ->
	     let id = nom_hyp h in
	     let c=val_hyp h in
	     let e=type_env_hyp h in
	     let ty = type_hyp h in
	     let s= sorte ty (type_of e  Evd.empty ty) in
	     let t = match s with
	       |"proposition"|"formule" ->
		  Tlist [Tterme (e,ty); Tsp; Tstr "(";pr_id id;Tstr ")"]
	       |"type"|"fonction"->
		  Tlist [pr_id id;Tsp;Tstr "un lment de";Tsp;Tterme (e,ty)]
	     in
	       lterm:=(!lterm)@[t];
	  )
          hyps in
    pr_list ",et" (fun x->x) (!lterm)

(**********************************************************************
 Recupere le script
*)

(***********************************************************************
Menage
*)
let rec menage t =
  let menagel l =
    List.filter (fun x -> match x with
		   | Trien -> false
		   | Tsp -> false
		   | Tnl -> false
		   | Tstr "" -> false
		   | Tlist [] -> false
		   | Top ("nl", []) -> false
		   | Tindent [] -> false
		   |_-> true)
      (List.map menage l)
  in
    match t with
      | Tlist l -> let l= (menagel l) in
	  (match l with
	     |[x] -> x
	     |_-> Tlist l)
      | Tindent l -> Tindent (menagel l)
      | Top ("enumerer" as s,l) -> Top (s, List.map menage l)
      | Top (s,l) -> Top (s, menagel l)
      |_-> t
(* devrait faire la meme chose que menage, mais non.
let rec menage1 t =
  let menagel l =
    List.filter (fun x -> match menage1 x with
		   | Trien -> false
		   | Tsp -> false
		   | Tnl -> false
		   | Tstr "" -> false
		   | Tlist [] -> false
		   | Top ("nl", []) -> false
		   | Tindent [] -> false
		   |_-> true)
      l
  in
    match t with
      | Tlist l -> Tlist (menagel l)
      | Tindent l -> Tindent (menagel l)
      | Top (s,l) -> Top (s, menagel l)
      |_-> t
*)

let rec applatit t =
  match t with
    | Tlist l -> let l = List.map applatit l in
	Tlist (List.fold_left (fun r x -> match x with
				 | Tlist l1 -> r @ l1
				 |_-> r@[x]) [] l)
    | Tindent l -> Tindent (List.map applatit l)
    | Top (s,l) -> Top (s, List.map applatit l)
    |_-> t

(* compacte les deductions successives #A_grace_a_#B [a ;b]
ex: d1= De_#A_on_deduit_#B
    d2= donc_#A
    d3= et_finalement_#A

    d1= On_obtient_#A
    d2= puis_#A
    d3= et_enfin_#A
*)
let comp_deduit1 d1 d2 d3 k l =
  let rec comp_deduit1 r r1 l =
    match l with
	[] -> if r1<>[] then r@[r1] else r
      | (Tph (Top ("#A_grce__#B",(Top(d,la))::g) as x) as y)::l 
      | ((Top ("#A_grce__#B",(Top(d,la))::g) as x) as y)::l ->
	  if d=d1
	  then comp_deduit1 r (r1@[x]) l
	  else comp_deduit1 (r@[r1]@[[y]]) [] l
      | (Ttac _ as x)::l->
	  if r1=[]
	  then comp_deduit1 (r@[[x]]) [] l
	  else comp_deduit1 r (r1@[x]) l
      |x::l -> comp_deduit1 (r@[r1]@[[x]]) [] l
  in
  let r = List.filter (fun x -> x<> []) (comp_deduit1 [] [] l) in
    List.map
      (fun l ->
	 match l with
	   | [Top ("#A_grce__#B",(Top(d,_))::_) as x] ->
	       if d=d1 then Tph x else x
	   | [x] -> x
	   |(Top ("#A_grce__#B",(Top(d,_))::_) as x)::l ->
	      if d=d1
	      then (let n= List.length (List.filter (fun x -> match x with
						       |Top _ -> true
						       |_->false)
					  l) in
		    let c=ref 0 in
		      Tph
			(Tlist
			   ((Top ("#A,",[x]))
			    ::(List.map
				 (fun y -> match y with
				    | Top ("#A_grce__#B",(Top(d,la))::g)->
					c:=(!c)+1;
					let y = if !c=n && n>2
					then Top ("#A_grce__#B",(Top (d3,ntl la k))::g)
					else Top ("#A_grce__#B",(Top (d2,ntl la k))::g)
					in if !c<n then Top ("#A,",[y]) else y
				    |_-> y)
				 l))))
	      else Tlist (x::l)
	   | _-> Tlist l)
      r
      
let comp_deduit d1 d2 d3 k t =
  let rec comp_deduit t =
    match t with
      | Tph p -> Tph (comp_deduit p)
      | Tlist l -> let l = List.map comp_deduit l in Tlist (comp_deduit1 d1 d2 d3 k l)
      | Tindent l -> let l = List.map comp_deduit l in Tindent (comp_deduit1 d1 d2 d3 k l)
      | Top (s,l) ->  let l = List.map comp_deduit l in Top (s, (comp_deduit1 d1 d2 d3 k l))
      |_-> t
  in comp_deduit t

(***********************************************************************
  Traduction en differentes langues
*)

let langue = ref "en";;

let traductions=[
("Dmontrons_#A_:",["Let us prove_#A_:"]);
("_prouver:_#A",["To_prove:_#A"]);
("#A_est_clair",["#A_is_clear"]);
("donc_#A_,",["hence_#A_,"]);
("Donc_#A_,",["Hence_#A_,"]);
("donc_#A",["hence_#A"]);
("et_finalement_#A",["and_finally_#A"]);
("puis_#A",["then_#A"]);
("et_enfin_#A",["and_finally_#A"]);
("De_#A_on_dduit_#B",["From_#A_we_deduce_#B"]);
("De_#A_:_#B_on_dduit_#C",["From_#A_:_#B_we_deduce_#C"]);
("Ainsi_#A",["Hence_#A"]);
("Ainsi_,_#A_,_#B",["Hence_,_#A_,_#B"]);
("De_plus_#A",["More_#A"]);
("Pour_prouver_#A_,_#B_procdons_ainsi:",["To_prove_#A_,_#B_we_follow_this_way:"]);
("#A_grce__#B",["#A_by_#B"]);
("On_a_#B",["We_have_#A"]);
("Par_#A_on_a_#B",["By_#A_we_have_#B"]);
("Distinguons_#N_cas_pour_#A",["We_distinguish_#N_cases_for_#A"]);
("Par_rcurrence_sur_la_dfinition_de_#A",["By_induction_on_the_definition_of_#A"]);
("Ainsi,_dans_tous_les_cas_on_a_montr_#A",["Hence,_for_every_case_we_have_proved_#A"]);
("Ainsi,_dans_tous_les_cas,_de_#A_,_on_dduit_#B",["Hence,_for_every_case,_from_#A_,_we_deduce_#B"]);
("De_la_dfinition_de_#A_,_de_#B_,_et_de_#C_,_on_dduit_#D",
  ["From_the_definition_of_#A_,_from_#B_,_and_from_#C_,_we_deduce_#D"]);
("par_dfinition_de_#A",["by_definition_of_#A"]);
("Procdons_par_rcurrence_sur_#A_(de_type_#B_)",["We_proceed_by_induction_on_#A_(of_type_#B_)"]);
("- Cas de rcurrence",["- Inductive step"]);
("- Cas de base",["- Base case"]);
("Par_#A",["By_#A"]);
("grce__#A",["by_#A"]);
("en_choisissant_#A",["choosing_#A"]);
("par_l'galit_#A",["by_equality_#A"]);
("on_obtient_#A",["we_get_#A"]);
("On_obtient_#A",["We_get_#A"]);
("on_peut_supposer_#A",["we_can_suppose_#A"]);
("Soit_#A",["Let_#A"]);
("On_a_#A",["We_have_#A"]);
("_partir_de_#A_:_#B",["From_#A_:_#B"]);
("_partir_de_#A_:_#B_on_dduit_#C",["From_#A_:_#B_we_deduce_#C"]);
("On_obtient_une_contradiction_avec_#A_et_#B",["We_get_a_contradiction_with_#A_and_#B"]);
("constructeur_#A",["constructor_#A"]);
("Avec_#A_:_#B_on_obtient_une_contradiction",["With_#A_:_#B_we_get_a_contradiction"]);
("De_#A_et_#B_:_#C_on_obtient_une_contradiction",["From_#A_and_#B_:_#C_we_get_a_contradiction"]);
("Soient ",["Let "]);
("Soit ",["Let "]);
(" des propositions",[" be_propositions"]);
(" une proposition",[" be a proposition"]);
(" telles que ",[" such that "]);
(" telle que ",[" such that "]);
(" tels que ",[" such that "]);
(" tel que ",[" such that "]);
(" des lments de ",[" be elements of "]);
("des lments de",["be elements of"]);
(" un lment de ",[" be an element of "]);
("un lment de",["be an element of"]);
("Supposons",["Suppose"]);
("et",["and"]);
("une contradiction",["a contradiction"]);
("Thorme",["Theorem"]);
("Dmonstration",["Proof"]);
("CQFD",["QED"]);
];;

let traduit s =
  try (let l = List.assoc s traductions in
	 match !langue with
	   |"fr"->s
	   |"en"->List.hd l
	   |_->s
      )
  with _-> s (* "##"^s^"##" *)

(***********************************************************************
  texte vers output coq
*)

(*
 Substitution des variables d'un texte
*)
let split c s =
  let l = ref [] in
  let i = ref 0 in
  let n= String.length s in
    while (!i)<n do
      try (let j = String.index_from s !i c in
	     l:=(!l)@[String.sub s !i (j-(!i))];
	     i:=j+1)
      with _->l:=(!l)@[String.sub s !i (n-(!i))];
	i:=n
    done;
    !l

let substitue_texte s l =
  let s = traduit s in
  let s = split '_' s in
  let r = ref (str "") in
  let l1 = ref l in
    List.iter (fun x -> if List.mem x ["#A";"#B";"#C";"#D";"#E";"#N"]
	       then (let a::l2 = !l1 in
		       r:=(!r)++ spc () ++ a;
		       l1:=l2)
	       else (match x with
		       |"R" -> r:=(!r)++ spc () ++ fnl()
		       |_-> r:=(!r)++ spc () ++ str x))
      s;
    !r
	

let rec to_coq t =
  match t with
    |Top("hypothese",l) -> to_coq (Tlist l)
    |Ttac lt -> str ""
       (* prl_coq (List.map
			   (fun (r,i) ->
			      Tstr ("#"^(string_of_rule r)^
				    ".("^(string_of_int i)^")#"))
			   lt)
       *)
    |Top ("numrer",l) ->
       let c = ref 0 in
	 hov 0 (List.fold_left (fun r x -> c:=(!c)+1;
				  r++
				  hov 0 (str ((string_of_int !c)^") ")++
					 (to_coq x)) ++
				  Pp.cut()) (str "") l)
    |Tph a -> hov 0 ((to_coq a)++str".")
    |Top ("#A,",[x])-> hov 0 ((to_coq x)++str ", "++Pp.cut())
    |Top (",et",l)->
       (match l with
	  |[] -> str ""
	  |[x]-> to_coq x
	  |x::l -> let y::l = List.rev l in
	    let l=List.rev l in
	      hov 0 ((to_coq x) ++
		     (List.fold_left (fun r x -> r++ str "," ++ spc() ++ (to_coq x))  (str "") l)++
		     spc ()++ str (traduit (t_et ()))++spc() ++ (to_coq y)++spc()))
    |Top (",",l)->
       (match l with
	  |[] -> str ""
	  |x::l -> hov 0 ((to_coq x) ++
			  (List.fold_left (fun r x -> r++ str "," ++ spc() ++ (to_coq x))  (str "") l)++
			  spc()))
    |Top ("#A_grce__#B",x::l) ->
       (match l with
	  | [] -> to_coq x
	  |_-> hov 0 ((to_coq x) ++ Pp.cut()++(hov 0 (str "  " ++ (prl_coq l))) ))
    |Trien -> str "Trivial."
    |Tsp -> spc ()
    |Tnl -> fnl ()
    |Tstr s -> str (traduit s)
    |Tnom s -> str s
    |Tlist l -> prl_coq l
    |Tseq s -> str "APROUVER**"
    |Tterme (e,c)-> prterm_env_at_top e c
    |Tindent l -> hov 0 (str "   "++ (prl_coq l)++ fnl ())
    |Top (s,l)->
       if s="nl"
       then  prl_coq l
       else  hov 0 (substitue_texte s (List.map to_coq l)) (* prl_coq_s s l*)

and prl_coq l =
  match l with
    |[]->str ""
    |a::l-> hov 0 ((List.fold_left (fun r x -> r++ Pp.cut() ++ (to_coq x))  (to_coq a) l)++
		   Pp.cut())

and prl_coq_s s l = hov 0 (List.fold_left (fun r x -> r++Pp.cut()++ (to_coq x))  (str s) l)

(***********************************************************************
*)
      
let h_ul l =
  (List.fold_left (fun r x -> r^"\n<LI>"^x) "<UL>" l)^"\n</UL>"


let rec to_html t =
  match t with
    |Top("hypothese",l)-> to_html (Tlist l)
    |Top ("numrer",l) ->
       let c = ref 0 in
	 h_ul (List.map (fun x -> c:=(!c)+1;
			   (string_of_int !c)^") "^(to_html x)) l)
    |Tph a -> (to_html a)^".<BR>\n"
    |Top ("#A,",[x])-> (to_html x)^", <BR>\n"
    |Top (",et",l)->
       (match l with
	  |[] -> ""
	  |[x]-> to_html x
	  |x::l ->
	     let y::l = List.rev l in
	     let l=List.rev l in
	       (to_html x)^
	       (List.fold_left (fun r x -> r^", "^(to_html x))  "" l)^
	       " "^(traduit (t_et ()))^" "^(to_html y)^" ")
    |Top (",etnl",l)->
       (match l with
	  |[] -> ""
	  |[x]-> to_html x
	  |x::l ->
	     let y::l = List.rev l in
	     let l=List.rev l in
	       (to_html x)^
	       (List.fold_left (fun r x -> r^",<BR>\n "^(to_html x))  "" l)^
	       "<BR>\n "^(traduit (t_et ()))^" "^(to_html y))
    |Top (",",l)->
       (match l with
	  |[] -> ""
	  |x::l -> (to_html x)^
	     (List.fold_left (fun r x -> r^", "^(to_html x))  "" l)^" ")
    |Top ("#A_grce__#B",x::l) ->
       (match l with
	  | [] -> to_html x
	  |_-> (to_html x)^"<IMG SRC=\"grace_a.gif\" ALT=\""^
	     (string_of_stream (prl_coq l))^"\" WIDTH=10 HEIGHT=10>")
    |Ttac lt ->
       "<IMG SRC=\"tactic.gif\" ALT=\""^
       (List.fold_left (fun s (r,i) ->
			  s^(string_of_rule r)^
			  ".("^(string_of_int i)^")")
	  ""
	  lt)^"\" WIDTH=15 HEIGHT=15>"
    |Trien -> "Trivial.<BR>\n"
    |Tsp -> " "
    |Tnl -> "\n"
    |Tstr s -> (traduit s)
    |Tnom s -> "<FONT color=red><I>"^s^"</I></FONT>"
    |Tlist l -> prl_html l
    |Tseq s -> "<B>"^"APROUVER***"^"</B>"
    |Tterme (e,c)-> if hd_is_mind e c "False"
      then (traduit "une contradiction")
      else "<I>"^(string_of_constr e c)^"</I>"
    |Tindent l -> "<BLOCKQUOTE>"^(prl_html l)^"</BLOCKQUOTE>"
    |Top (s,l)->
       if s="nl"
       then  prl_html l
       else  (substitue_texte_html s (List.map to_html l)) (* prl_html_s s l*)

and prl_html l =
  match l with
    |[]-> ""
    |a::l-> (List.fold_left (fun r x -> r^(to_html x))  (to_html a) l)

and prl_html_s s l = (List.fold_left (fun r x -> r^(to_html x))  s l)

and substitue_texte_html s l =
  let s = traduit s in
  let s = split '_' s in
  let r = ref "" in
  let l1 = ref l in
    List.iter (fun x -> if List.mem x ["#A";"#B";"#C";"#D";"#E";"#N"]
	       then (let a::l2 = !l1 in
		       r:=(!r)^" "^a;
		       l1:=l2)
	       else (match x with
		       |"R" -> r:=(!r)^" \n"
		       |_-> r:=(!r)^" "^x))
      s;
    !r	  
(***********************************************************************
XML
*)
(* change les < > *)

let string_subst s c d =
  let rec subst s =
    try (let i = String.index s c in
	 let s1 = String.sub s 0 i in
	 let s2 = subst (try (String.sub s (i+1) ((String.length s)-(i+1))) with _ -> "") in
	   s1^d^s2)
    with _ -> s
  in subst s

let xml s =
  let s = string_subst s '<' "&lt;" in
  let s = string_subst s '>' "&gt;" in
  let s = string_subst s '' "&#233;" in
  let s = string_subst s '' "&#224;" in
  let s = string_subst s '' "&#226;" in
  let s = string_subst s '' "&#192;" in
    s

let substitue_texte_xml s l =
  let s = split '_' (xml s) in
  let r = ref "" in
  let l1 = ref l in
    List.iter (fun x -> if List.mem x ["#A";"#B";"#C";"#D";"#E";"#N"]
	       then (let a::l2 = !l1 in
		       r:=(!r)^" "^a;
		       l1:=l2)
	       else (match x with
		       |"R" -> r:=(!r)^" \n"
		       |_-> r:=(!r)^" "^x))
      s;
    !r	  
let sans_accent s =
  let s = string_subst s '' "e" in
  let s = string_subst s '' "a" in
  let s = string_subst s '' "a" in
  let s = string_subst s '' "A" in
  let s = string_subst s ':' "" in
let s = string_subst s '(' "" in
let s = string_subst s ')' "" in
  let s = string_subst s ',' "" in
  let s = string_subst s '#' "ARG" in
    s

(* Eviter de passer par un fichier! *)
let terme_xml e c =
  try (Xml.pp (ProofTree2Xml.constr_to_xml c Evd.empty e) (Some "bidon");
       let ic = open_in "bidon.xml" in
       let s = ref "" in
	 try (while true do
		s :=(!s)^(input_line ic);
	      done;
	      !s)
	 with _ -> !s)
  with _-> print_string ("PROBLEME: "^(string_of_constr e c)^"\n");
    (xml (string_of_constr e c))

(* La table des termes rencontres *)
let les_termes = ref (Hashtbl.create 51);;
let ntermes = ref 0

let numero_terme t =
   try (Hashtbl.find !les_termes t)
   with _ -> (ntermes:=(!ntermes)+1;
              Hashtbl.add !les_termes t !ntermes;
              !ntermes)

let rec to_xml t =
  match t with
    |Top("hypothese",[t;_; h; _]) ->
       "<HYPOTHESE>"^(to_xml t)^(to_xml h)^"</HYPOTHESE>"
    |Top ("numrer",l) ->
       (List.fold_left (fun r x -> r^"\n"^(to_xml x)) "<ENUMERER>" l)^"</ENUMERER>"
    |Tph a -> "<PHRASE>"^(to_xml a)^"</PHRASE>"
    |Top ("#A,",[x])-> "<VIRGULE>"^(to_xml x)^"</VIRGULE>"
    |Top (",et",l)->
       (List.fold_left (fun r x -> r^"\n"^(to_xml x)) "<LISTE_VIRGULE_ET>" l)^"</LISTE_VIRGULE_ET>"
    |Top (",etnl",l)->
       (List.fold_left (fun r x -> r^"\n"^(to_xml x)) "<LISTE_VIRGULE_ET_NL>" l)^"</LISTE_VIRGULE_ET_NL>"
    |Top (",",l)->
       (List.fold_left (fun r x -> r^"\n"^(to_xml x)) "<LISTE_VIRGULE>" l)^"</LISTE_VIRGULE>"
    |Top ("#A_grce__#B",x::l) ->
      ( match l with
	 |[] -> (to_xml x)
	 |_ ->
	    "<GRACE_A>"^(List.fold_left (fun r x -> r^"\n"^(to_xml x)) (to_xml x) l)^"</GRACE_A>")
    |Ttac lt ->
       "<TACTICS>"^
       (List.fold_left (fun s (r,i) ->
			  s^"<TACTIC RANK=\""^(string_of_int i)^"\">"^
			  (xml (string_of_rule r))^". </TACTIC>")
	  ""
	  lt)^"</TACTICS>"
    |Trien -> "<TEXTE>Trivial.</TEXTE>"
    |Tsp -> " " (* corriger *)
    |Tnl -> "" (* corriger *)
    |Tstr s -> "<TEXTE>"^(xml s)^"</TEXTE>"
    |Tnom s -> "<IDENT>"^(xml s)^"</IDENT>"
    |Tlist l -> "<LISTE>"^(prl_xml l)^"</LISTE>"
    |Tseq s -> "<A_PROUVER>"^"APROUVER****"^"</A_PROUVER>"
    |Tterme (e,c)-> let n = string_of_int (numero_terme c) in
      if hd_is_mind e c "False"
      then "<TEXTE>"^(traduit "une contradiction")^"</TEXTE>"
      else "<TERME ID=\""^n^"\">"^(terme_xml e c)^"</TERME>"
    |Tindent l -> "<DEMONSTRATION>"^(prl_xml l)^"</DEMONSTRATION>"
    |Top (s,l)->
       if s="nl"
       then  prl_xml l
(*       else  "<TEXTE>"^(substitue_texte_xml s (List.map to_xml l))^"</TEXTE>" *)
       else  let s = sans_accent s in
	 "<"^s^">"^(prl_xml l)^"</"^s^">"

and prl_xml l =
  match l with
     |[]-> ""
    |a::l-> (List.fold_left (fun r x -> r^(to_xml x))  (to_xml a) l)


(***********************************************************************
Tests
*)

let get_pf () =
  let pf = (Tacmach.proof_of_pftreestate (Pfedit.get_pftreestate())) in
    pf
      
let exp_coq p =
  pp (to_coq p)

let exp_html pf p =
  let h=to_html p in
  let chan_out = open_out "explication.html" in
    output_string chan_out
      ("<B>"^(traduit "Thorme")^": "^(to_html (Tterme (env pf, concl pf)))^
       "<BR>"^(traduit "Dmonstration")^":<BR></B>\n"^
       h
       ^"<BR>"^(traduit "CQFD"));
    close_out  chan_out

let exp_xml pf p =
  ntermes:=0;
  les_termes:=Hashtbl.create 51;
  let h=to_xml p in
  let chan_out = open_out "explication.xml" in
    output_string chan_out
      ("<TEXTE_DE_PREUVE>"^
       "<THEOREME>"^(to_xml (Tterme (env pf, concl pf)))^"</THEOREME>"^
       "<DEMONSTRATION>"^
       h^
       "</DEMONSTRATION>"^
       "</TEXTE_DE_PREUVE>");
    close_out  chan_out



let exp g t =
  let g0= (!p_grace_a) in
  let t0=(!p_type) in
    p_grace_a:=g;
    p_type:=t;
    Format.set_max_boxes 500;
    (* Format.set_margin 120;*)
    let pf = preuve_courante true in
    let p = expliquer0 (virer_intros true pf) in
    let p = applatit (menage p) in
    let p = comp_deduit "De_#A_on_dduit_#B" "donc_#A" "et_finalement_#A" 1 p in
    let p = comp_deduit "On_obtient_#A" "puis_#A" "et_enfin_#A" 0 p in
      p_grace_a:=g0;
      p_type:=t0;
      p


let qed_exp ()=
  let pf = preuve_courante true in
  let e = exp true true in
    exp_coq e;
    exp_html pf e;
    exp_xml pf e;
    let namef = (get_current_proof_name ()) in
    let env = Global.env() in
    let current_sign = Global.named_context() in
    let _,(rconst,strength,_) = cook_proof () in
    let {const_entry_body=r;const_entry_type = tr} = rconst in
      delete_current_proof ();
      let rr= r in
      let cd = Entries.DefinitionEntry
		 {const_entry_body=rr;
		  const_entry_type=tr;
		  const_entry_opaque = true} in
      let sp = Declare.declare_constant namef (cd,IsProof Lemma) in
      let s = constr_of_reference (ConstRef (snd sp)) in
	message "\n";
	message ((string_of_id namef)^" is defined.");
	
open Vernacextend;;

(* le texte d'explication de la preuve courante *)
VERNAC COMMAND EXTEND Exp
  [ "Exp" ] -> [ exp_coq (exp true true)]
    END	 
  
VERNAC COMMAND EXTEND Exp1
  [ "Exp1" ] -> [ exp_coq (exp false false) ]
    END	   

(* la preuve avec des sequents *)
VERNAC COMMAND EXTEND Exp0
  [ "Exp0" ] -> [ pp (print_proof Evd.empty empty_named_context (get_pf ()))]
    END	   

(* affiche le texte d'explication de la preuve avant de faire Qed *)
VERNAC COMMAND EXTEND QedExp
  [ "QedExp" ] -> [ qed_exp ()]
    END	   

(* en anglais *)
VERNAC COMMAND EXTEND SetExpEn
  [ "SetExpEn" ] -> [ langue:="en"]
    END	 
(* en anglais *)
VERNAC COMMAND EXTEND SetExpFr
  [ "SetExpFr" ] -> [ langue:="fr"]
    END	 

(* Avec Claudio ************************************************************)

open Xml

let terme_xml2 e c =
  try (ProofTree2Xml.constr_to_xml c Evd.empty e)
  with _-> print_string ("PROBLEME: "^(string_of_constr e c)^"\n");
    xml_cdata (xml (string_of_constr e c))

let rec to_xml2 t =
  match t with
    |Top("hypothese",[t;_; h; _]) ->
       xml_nempty "HYPOTHESE" [] [< to_xml2 t; to_xml2 h>]
    |Top ("numrer",l) ->
       xml_nempty "ENUMERER" []
	(List.fold_left (fun r x -> [<r; to_xml2 x>]) [<>] l)
    |Tph a -> xml_nempty "PHRASE" [] (to_xml2 a)
    |Top ("#A,",[x])-> xml_nempty "VIRGULE" [] (to_xml2 x)
    |Top (",et",l)->
       xml_nempty "LISTE_VIRGULE_ET" []
	(List.fold_left (fun r x -> [<r;to_xml2 x>]) [<>] l)
    |Top (",etnl",l)->
       xml_nempty "LISTE_VIRGULE_ET_NL" []
	(List.fold_left (fun r x -> [<r;to_xml2 x>]) [<>] l)
    |Top (",",l)->
       xml_nempty "LISTE_VIRGULE" []
	(List.fold_left (fun r x -> [<r;to_xml2 x>]) [<>] l)
    |Top ("#A_grce__#B",x::l) ->
      ( match l with
	 |[] -> (to_xml2 x)
	 |_ ->
	     xml_nempty "GRACE_A" []
	    (List.fold_left (fun r x -> [<r;to_xml2 x>]) (to_xml2 x) l))
    |Ttac lt ->
       xml_nempty "TACTICS" []
       (List.fold_left (fun s (r,i) ->
			  [<s;xml_nempty "TACTIC" ["RANK",string_of_int i]
			      (xml_cdata (xml (string_of_rule r)))>])
	  [<>]
	  lt)
    |Trien -> xml_nempty "TEXTE" [] (xml_cdata "Trivial.")
    |Tsp -> [<>] (* corriger *)
    |Tnl -> [<>] (* corriger *)
    |Tstr s ->  xml_nempty "TEXTE" [] (xml_cdata (xml s))
    |Tnom s -> xml_nempty "IDENT" [] (xml_cdata (xml s))
    |Tlist l -> xml_nempty  "LISTE" [] (prl_xml2 l)
    |Tseq s -> xml_nempty "A_PROUVER" [] (xml_cdata "DA DIMONSTRARE****")
    |Tterme (e,c)-> let n = string_of_int (numero_terme c) in
      if hd_is_mind e c "False"
      then xml_nempty "TEXTE" [] (xml_cdata (traduit "une contradiction"))
      else xml_nempty "TERME" ["ID",n]  (terme_xml2 e c)
    |Tindent l -> xml_nempty "DEMONSTRATION" [] (prl_xml2 l)
    |Top (s,l)->
       if s="nl"
       then  prl_xml2 l
       else  let s = sans_accent s in
	 xml_nempty s [] (prl_xml2 l) 

and prl_xml2 l =
  match l with
     |[]-> [<>]
    |a::l-> (List.fold_left (fun r x -> [<r;to_xml2 x>])  (to_xml2 a) l)

let exp_xml_stream () =
  let pf = preuve_courante true in
  let p = exp true true in
    ntermes:=0;
  les_termes:=Hashtbl.create 51;
  xml_nempty "TEXTE_DE_PREUVE" []
    [< xml_nempty "THEOREME" []
	 (to_xml2 (Tterme (env pf, concl pf)));
       xml_nempty "DEMONSTRATION" [] (to_xml2 p) >]
    
(*************************************************************)

(* la preuve xml *)
VERNAC COMMAND EXTEND Exp2
  [ "Exp2" ] -> [ pp (exp_xml_stream ()) None]
    END	   

(*
 ~/pcoq/V7/bin/coqtop.byte.exe
Declare ML Module "arbre_de_preuve".
Declare ML Module "explication".
 Goal (a,b:nat)(b:bool)b=b.
Intros a b n.
Exp.

 Lemma l1:(A,B,C:Prop)A/\(B\/C)->(A/\B)\/(A/\C).
Intros.
Elim H; Intros.
Elim H1; Intros.
Left.
Split; Auto.

Right.
Split; Auto.

Drop.
open Arbre_de_preuve;;
open Explication;;
#use "base_include";;
let u=expliquer0 (virer_intros true (expanser_tactiquelles (get_pf ())));;
menage u;;



QedExp.

Goal (A,B:Prop)A->B->A.
Intros.

Drop.
open Pp;;
let a = hov 0 (str "aaa"++spc()++str"bbbbb");;

pp_with  Format.str_formatter a;;
let s = Format.flush_str_formatter ();;

*)
